# -*- coding: UTF-8 -*-
#
# Copyright 2019-2022 Flávio Gonçalves Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import process
import click
from click.core import Command, Context, Group, HelpFormatter
import os
from typing import Any, Callable, Dict, List, Optional, Sequence, Union
import sys


__multi_command__ = None


class TaskioTestContext(Context):

    def __init__(
            self,
            command: Command,
            parent: Optional[Context] = None,
            info_name: Optional[str] = None,
            obj: Optional[Any] = None,
            auto_envvar_prefix: Optional[str] = None,
            default_map: Optional[Dict[str, Any]] = None,
            terminal_width: Optional[int] = None,
            max_content_width: Optional[int] = None,
            resilient_parsing: bool = False,
            allow_extra_args: Optional[bool] = None,
            allow_interspersed_args: Optional[bool] = None,
            ignore_unknown_options: Optional[bool] = None,
            help_option_names: Optional[List[str]] = None,
            token_normalize_func: Optional[Callable[[str], str]] = None,
            color: Optional[bool] = None,
            show_default: Optional[bool] = None,
    ) -> None:
        super().__init__(command, parent, info_name, obj, auto_envvar_prefix,
                         default_map, terminal_width, max_content_width,
                         resilient_parsing, allow_extra_args,
                         allow_interspersed_args, ignore_unknown_options,
                         help_option_names, token_normalize_func, color,
                         show_default)


class TaskioMultiCommand(click.MultiCommand):

    def __init__(
            self, name=None, invoke_without_command: bool = False,
            no_args_is_help=None, subcommand_metavar=None,
            chain: bool = False, result_callback=None, **attrs,
    ) -> None:
        self.conf = {}
        if "taskio_conf" in attrs:
            self.conf = attrs.pop("taskio_conf")
            sys.modules[__name__].__multi_command__ = self
        self.loader = process.TaskioLoader(self.conf, self)
        self.loader.load()
        if "taskio_conf" in attrs:
            if self.loader.full_name:
                name = self.loader.full_name
        super().__init__(name, invoke_without_command, no_args_is_help,
                         subcommand_metavar, chain, result_callback, **attrs)
        self.context_class = TaskioTestContext

    def list_commands(self, ctx: Any) -> List[str]:
        rv = []
        groups = []
        for source in self.loader.sources:
            for key, item in source.__dict__.items():
                if isinstance(item, Command):
                    if isinstance(item, Group):
                        groups.append(item)
                    rv.append(item.name)
        for group in groups:
            for key, item in group.commands.items():
                if item.name in rv:
                    rv.remove(item.name)
        rv.sort()
        return rv

    def get_command(self, ctx: Any, cmd_name: str) -> Optional[click.Command]:
        for source in self.loader.sources:
            for name, val in source.__dict__.items():
                if isinstance(val, Command) and val.name == cmd_name:
                    return val
        return

    def format_usage(self, ctx: Context, formatter: HelpFormatter) -> None:
        pieces = self.collect_usage_pieces(ctx)
        formatter.write(self.loader.full_name)
        formatter.write("\n")
        formatter.write_usage(self.loader.name, " ".join(pieces))


class TaskioGroup(Group):

    def __init__(
        self,
        name: Optional[str] = None,
        commands: Optional[Union[Dict[str, Command],
                                 Sequence[Command]]] = None,
        **attrs: Any,
    ) -> None:
        super().__init__(name, commands, **attrs)
        self._multi_command = sys.modules[__name__].__multi_command__

    def format_usage(self, ctx: Context, formatter: HelpFormatter) -> None:
        pieces = self.collect_usage_pieces(ctx)
        formatter.write(self._multi_command.loader.full_name)
        formatter.write("\n")
        prog = "%s %s" % (self._multi_command.loader.name,
                          self.resolve_params(ctx))
        formatter.write_usage(prog, " ".join(pieces))

    def resolve_params(self, ctx):
        params = ""
        current_context = ctx
        while current_context:
            if current_context.parent is not None:
                params = " ".join([current_context.info_name, params])
            current_context = current_context.parent
        return params

    def group(
        self, *args: Any, **kwargs: Any
    ) -> Callable[[Callable[..., Any]], "Group"]:
        """A shortcut decorator for declaring and attaching a group to
        the group. This takes the same arguments as :func:`group` and
        immediately registers the created group with this group by
        calling :meth:`add_command`.

        To customize the group class used, set the :attr:`group_class`
        attribute.

        .. versionchanged:: 8.0
            Added the :attr:`group_class` attribute.
        """
        from .decorators import group

        if self.group_class is not None and "cls" not in kwargs:
            if self.group_class is type:
                kwargs["cls"] = type(self)
            else:
                kwargs["cls"] = self.group_class

        def decorator(f: Callable[..., Any]) -> "TaskioGroup":
            cmd = group(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator


class TaskioCommand(Command):

    def format_usage(self, ctx: Context, formatter: HelpFormatter) -> None:
        pieces = self.collect_usage_pieces(ctx)
        formatter.write(self.loader.full_name)
        formatter.write("\n")
        formatter.write_usage(self.loader.name, " ".join(pieces))


class TaskioContext(object):

    def __init__(self, **kwargs):
        self.current = os.getcwd()


def get_context():
    return click.make_pass_decorator(TaskioContext, ensure=True)



# -*- coding: UTF-8 -*-
#
# Copyright 2019-2022 Flávio Gonçalves Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import TaskioCommand, TaskioMultiCommand, TaskioGroup
from click.decorators import _make_command
import typing as t

F = t.TypeVar("F", bound=t.Callable[..., t.Any])
FC = t.TypeVar("FC", bound=t.Union[t.Callable[..., t.Any], TaskioCommand])


def group(name: t.Optional[str] = None,
          **attrs: t.Any) -> t.Callable[[F], TaskioGroup]:
    """Creates a new :class:`Group` with a function as callback.  This
    works otherwise the same as :func:`command` just that the `cls`
    parameter is set to :class:`Group`.
    """
    attrs.setdefault("cls", TaskioGroup)
    return t.cast(TaskioGroup, command(name, **attrs))


def command(
    name: t.Optional[str] = None,
    cls: t.Optional[t.Type[TaskioCommand]] = None,
    **attrs: t.Any,
) -> t.Callable[[F], TaskioCommand]:
    r"""Creates a new :class:`Command` and uses the decorated function as
    callback.  This will also automatically attach all decorated
    :func:`option`\s and :func:`argument`\s as parameters to the command.

    The name of the command defaults to the name of the function with
    underscores replaced by dashes.  If you want to change that, you can
    pass the intended name as the first argument.

    All keyword arguments are forwarded to the underlying command class.

    Once decorated the function turns into a :class:`Command` instance
    that can be invoked as a command line utility or be attached to a
    command :class:`Group`.

    :param name: the name of the command.  This defaults to the function
                 name with underscores replaced by dashes.
    :param cls: the command class to instantiate.  This defaults to
                :class:`Command`.
    """
    if cls is None:
        cls = TaskioCommand

    def decorator(f: t.Callable[..., t.Any]) -> TaskioCommand:
        cmd = _make_command(f, name, attrs, cls)  # type: ignore
        cmd.__doc__ = f.__doc__
        return cmd

    return decorator


def root(
    name: t.Optional[str] = None,
    **attrs: t.Any,
) -> t.Callable[[F], TaskioCommand]:
    r"""Creates a new :class:`Command` and uses the decorated function as
    callback.  This will also automatically attach all decorated
    :func:`option`\s and :func:`argument`\s as parameters to the command.

    The name of the command defaults to the name of the function with
    underscores replaced by dashes.  If you want to change that, you can
    pass the intended name as the first argument.

    All keyword arguments are forwarded to the underlying command class.

    Once decorated the function turns into a :class:`Command` instance
    that can be invoked as a command line utility or be attached to a
    command :class:`Group`.

    :param name: the name of the command.  This defaults to the function
                 name with underscores replaced by dashes.
    :param cls: the command class to instantiate.  This defaults to
                :class:`Command`.
    """
    cls = TaskioMultiCommand

    def decorator(f: t.Callable[..., t.Any]) -> TaskioMultiCommand:
        cmd = _make_command(f, name, attrs, cls)  # type: ignore
        cmd.__doc__ = f.__doc__
        return cmd

    return decorator


# -*- coding: UTF-8 -*-
#
# Copyright 2019-2022 Flávio Gonçalves Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from gettext import gettext as _


class TaskioException(Exception):
    """ Base exception to be thrown by any part of the code and be displayed by
    the current runner.
    """

    def __init__(self, *args, **kwargs):
        super(TaskioException, self).__init__(*args, **kwargs)
        self._help = kwargs.get("help", None)
        self._source = kwargs.get("source", None)
        self._reason = kwargs.get("reason", None)
        self._show_usage = False

    @property
    def help(self):
        return self._help

    @help.setter
    def help(self, help):
        self._help = help

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        self._source = source

    @property
    def reason(self):
        return self._reason

    @reason.setter
    def reason(self, reason):
        self._reason = reason

    @property
    def show_usage(self):
        return self._show_usage

    @show_usage.setter
    def show_usage(self, show_usage):
        self._show_usage = show_usage


class TaskioArgumentError(TaskioException):
    """ An error thrown while parsing arguments with TaskioArgumentParser.
    """
    pass


class TaskioArgumentParser(argparse.ArgumentParser):
    """ Argument parser that trows an exception leaving stderr untouched.
    """

    def error(self, message):
        args = {'message': message}
        message = _("%(message)s") % args
        raise TaskioArgumentError(message)



# -*- coding: UTF-8 -*-
#
# Copyright 2019-2022 Flávio Gonçalves Garcia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .config import resolve_name, resolve_version
from cartola import sysexits
import importlib
import logging
import sys


logger = logging.getLogger(__name__)


class TaskioContext(object):

    def __init__(self, conf, **kwargs):
        self._conf = conf


class TaskioLoader(object):

    def __init__(self, conf, program=None, **kwargs):
        self._conf = conf
        self._root = kwargs.get("root", "taskio")
        self._program = program
        self._version = None
        self._name = None
        self._sources = []
        if not self._conf:
            print(
                "Taskio FATAL ERROR:\n Please provide a configuration to the "
                "command.")
            sys.exit(sysexits.EX_FATAL_ERROR)
        if not self._root or self._root not in self._conf:
            print("Taskio FATAL ERROR:\n  Please add a root to the command "
                  "configuration")
            sys.exit(sysexits.EX_FATAL_ERROR)

    def load(self):
        if "program" in self.conf:
            if "name" in self.conf['program']:
                self.conf['program']['name'] = resolve_name(
                    self.conf['program']['name']
                )
            if "version" in self.conf['program']:
                self.conf['program']['version'] = resolve_version(
                    self.conf['program']['version']
                )

        if "sources" in self.conf:
            for source in self.conf['sources']:
                self._sources.append(importlib.import_module(source))

    @property
    def program(self):
        return self._program

    @property
    def name(self):
        if "program" in self.conf and "name" in self.conf['program']:
            return self.conf['program']['name']
        return None

    @property
    def version(self):
        if "program" in self.conf and "version" in self.conf['program']:
            return self.conf['program']['version']
        return None

    @property
    def full_name(self):
        if self.name:
            name = self.name
            if self.version:
                return "%s %s" % (name, self.version)
            return name
        return None

    @property
    def conf(self):
        return self._conf[self._root]

    @property
    def sources(self):
        return self._sources
